import * as effector from 'effector';
import { StoreWritable } from 'effector';

type NonEmptyTuple<T = unknown> = [T, ...T[]];
type AnyObject = Record<string, unknown>;
type ValueOf<T> = T[keyof T];
declare const CONTAINER_STATUS: {
    readonly idle: "idle";
    readonly pending: "pending";
    readonly done: "done";
    readonly fail: "fail";
    readonly off: "off";
};
type ContainerStatus = ValueOf<typeof CONTAINER_STATUS>;
type StartResult<T> = Promise<{
    api: T;
}> | {
    api: T;
};
type EnableResult = Promise<boolean> | boolean;
type AnyAPI = AnyObject | null;
type AnyStartFn = (...x: any) => StartResult<AnyAPI>;
type AnyContainer = {
    id: string;
    domain: string;
    $status: StoreWritable<ContainerStatus>;
    start: AnyStartFn;
    dependsOn?: AnyContainer[];
    optionalDependsOn?: AnyContainer[];
    enable?: (..._: any) => EnableResult;
};
type AnyDeps = NonEmptyTuple<AnyContainer> | void;

type ValidateParams = Pick<AnyContainer, 'id' | 'domain' | 'dependsOn' | 'optionalDependsOn'>;
declare const ERROR: {
    readonly CONTAINER_ID_EMPTY_STRING: "Container ID cannot be an empty string.";
    readonly CONTAINER_DOMAIN_NAME_EMPTY_STRING: "Container Domain cannot be an empty string.";
    readonly depsIntersection: (intersection: string[], containerId: AnyContainer["id"]) => string;
};
type ContainerIdEmptyStringError = ValidateParams & {
    id: never;
    error: typeof ERROR.CONTAINER_ID_EMPTY_STRING;
};
type ContainerDomainNameEmptyStringError = ValidateParams & {
    domain: never;
    error: typeof ERROR.CONTAINER_DOMAIN_NAME_EMPTY_STRING;
};

type ExtractDeps<D extends AnyContainer[]> = {
    [K in D[number] as Awaited<ReturnType<K['start']>>['api'] extends Record<string, never> ? never : K['id']]: Awaited<ReturnType<K['start']>>['api'];
};
type Params<Id extends string, Domain extends string, API extends AnyAPI, Deps extends AnyDeps = void, OptionalDeps extends AnyDeps = void> = '' extends Id ? ContainerIdEmptyStringError : '' extends Domain ? ContainerDomainNameEmptyStringError : Deps extends void ? OptionalDeps extends void ? {
    id: Id;
    domain: Domain;
    start: () => StartResult<API>;
    enable?: () => EnableResult;
} : {
    id: Id;
    domain: Domain;
    optionalDependsOn: Exclude<OptionalDeps, void>;
    start: (_: void, optionalDeps: Partial<ExtractDeps<Exclude<OptionalDeps, void>>>) => StartResult<API>;
    enable?: (_: void, optionalDeps: Partial<ExtractDeps<Exclude<OptionalDeps, void>>>) => EnableResult;
} : OptionalDeps extends void ? {
    id: Id;
    domain: Domain;
    dependsOn: Exclude<Deps, void>;
    start: (deps: ExtractDeps<Exclude<Deps, void>>) => StartResult<API>;
    enable?: (deps: ExtractDeps<Exclude<Deps, void>>) => EnableResult;
} : {
    id: Id;
    domain: Domain;
    dependsOn: Exclude<Deps, void>;
    optionalDependsOn: Exclude<OptionalDeps, void>;
    start: (deps: ExtractDeps<Exclude<Deps, void>>, optionalDeps: Partial<ExtractDeps<Exclude<OptionalDeps, void>>>) => StartResult<API>;
    enable?: (deps: ExtractDeps<Exclude<Deps, void>>, optionalDeps: Partial<ExtractDeps<Exclude<OptionalDeps, void>>>) => EnableResult;
};
declare const createContainer: <Id extends string, Domain extends string, API extends AnyAPI, Deps extends AnyDeps = void, OptionalDeps extends AnyDeps = void>(__params: Params<Id, Domain, API, Deps, OptionalDeps>) => Exclude<Params<Id, Domain, API, Deps, OptionalDeps>, ContainerIdEmptyStringError | ContainerDomainNameEmptyStringError> & {
    $status: effector.StoreWritable<ContainerStatus>;
};

type ContainerId = AnyContainer['id'];
type ContainerDomain = AnyContainer['domain'];

type TransitiveDependency<Id extends string = ContainerId> = {
    id: Id;
    path: string;
};
type ContainersGraph = Record<ContainerId, {
    domain: ContainerDomain;
    strict: ContainerId[];
    optional: ContainerId[];
    transitive: {
        strict: TransitiveDependency<ContainerId>[];
        optional: TransitiveDependency<ContainerId>[];
    };
}>;
type DomainsGraph = Record<ContainerDomain, {
    containers: ContainerId[];
    strict: ContainerDomain[];
    optional: ContainerDomain[];
    transitive: {
        strict: TransitiveDependency<ContainerDomain>[];
        optional: TransitiveDependency<ContainerDomain>[];
    };
}>;

declare const compose: {
    up: <T extends AnyContainer[], C extends {
        apis?: boolean;
        debug?: boolean;
        onContainerFail?: (_: {
            id: AnyContainer["id"];
            error: Error;
        }) => unknown;
    }>(inputContainers: T, __config?: C) => Promise<undefined extends C ? {
        ok: boolean;
        data: {
            statuses: { [K in T[number]["id"]]: ContainerStatus; };
        };
    } : NonNullable<C>["apis"] extends true ? {
        ok: boolean;
        data: {
            apis: { [K_1 in T[number] as K_1["id"]]?: Awaited<ReturnType<K_1["start"]>>["api"] | undefined; };
            statuses: { [K in T[number]["id"]]: ContainerStatus; };
        };
    } : {
        ok: boolean;
        data: {
            statuses: { [K in T[number]["id"]]: ContainerStatus; };
        };
    }>;
    graph: <T extends "containers" | "domains" = "containers">(__containers: AnyContainer[], config?: {
        view?: T;
    }) => T extends "domains" ? {
        data: DomainsGraph;
        skippedContainers: {
            [x: string]: string[];
        };
        dependsOn: (_: AnyContainer["domain"][]) => DomainsGraph;
        requiredBy: (_: AnyContainer["domain"][]) => DomainsGraph;
    } : {
        data: ContainersGraph;
        skippedContainers: {
            [x: string]: string[];
        };
        dependsOn: (_: AnyContainer[]) => ContainersGraph;
        requiredBy: (_: AnyContainer[]) => ContainersGraph;
    };
};

export { compose, createContainer };
